# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
BAD_INPUTS = {
    "test_excess_mt.imcnp",
    "test_vol_redundant.imcnp",
    "testVerticalMode.imcnp",
    "test_missing_mat_for_mt.imcnp",
    "test_imp_redundant.imcnp",
    "test_broken_cell_surf_link.imcnp",
    "test_broken_surf_link.imcnp",
    "test_broken_transform_link.imcnp",
    "test_broken_mat_link.imcnp",
    "test_broken_complement.imcnp",
    "test_bad_syntax.imcnp",
    "test_extra_params.imcnp",
    "test_extra_data_param.imcnp",
    "test_extra_data_imp.imcnp",
    "number_conflict_pin_cell.imcnp",
}
IGNORE_FILES = {
    "testReadRec1.imcnp",
    "testReadRec2.imcnp",
    "testReadRec3.imcnp",
    "testReadTarget.imcnp",
    "bad_encoding.imcnp",
    "unicode.imcnp",
}

BAD_ENCODING_FILES = {
    "bad_encoding.imcnp",
    "unicode.imcnp",
}
